/*
 *	anysurrect.c
 *      CopyRight (C) 2006, Nikolaj Krivchenkov aka unDEFER <undefer@gmail.com>
 */

#define _LARGEFILE64_SOURCE

#include "any.h"
#include "super.h"

#if defined(_WIN32)
#include <windows.h>
#include <winioctl.h>

#define O_LARGEFILE	0
#endif

#include <sys/types.h>
#if defined(__linux__)
#include <sys/wait.h>
#define	MKDIR(a, b) mkdir(a, b)
#else
#include "misc.h"
#define	MKDIR(a, b) mkdir(a)
#define	ftruncate64(a, b) ftruncate(a, b)
#endif
#include <unistd.h>

#include <stdio.h>
#if defined(__linux__)
#include <stdio_ext.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#if defined(__linux__)
#include <sys/ioctl.h>
#include <linux/fd.h>

#include <dlfcn.h>
#endif

#include <assert.h>

#include <bitops.h>
#if defined(__linux__)
#include <endian.h>
#else
#include <sys/param.h>
#include "ext2_bitops.h"
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "anysurrect.h"

#include "new_inode.h"
#include "block_map.h"
#include "progress.h"
#include "version.h"

#include "mainform.h"

#include <QFileIconProvider>
#include <QMessageBox>
#include <locale.h>

#if defined(__linux__)
    
typedef void* dl_handle_t;
#define DLOPEN(dll) dlopen(dll, RTLD_LAZY | RTLD_GLOBAL)
#define DLSYM(handle, sym) dlsym(handle, sym)
#define DLSUF ".so"
#define DLCLOSE(handle) dlclose(handle)
    
#elif defined(_WIN32)

typedef HMODULE dl_handle_t;
#define DLOPEN(dll) LoadLibrary(dll)
#define DLSYM(handle, sym) GetProcAddress((HINSTANCE__*) handle, sym)
#define DLSUF ".dll"
#define DLCLOSE(handle) FreeLibrary(handle)
#define dlerror()	"UNKNOWN ERROR"

#endif


#define BLKGETSIZE _IO(0x12,96) /* return device size /512 (long *arg) */

/*
------------------------------------8<----------------------------------
--------------------------->8-------------------------------------------
 */

static unsigned long *SKIP_TO_BLOCK = NULL;

void anysurrect_fromblock(struct any_sb_info *info)
{
	char *mes;
	any_size_t max_size=1;

	int type;

	info_for_anysurrect_frags_list = info;

	if (!SKIP_TO_BLOCK) 
	{
		SKIP_TO_BLOCK = 
			(unsigned long *)
			calloc(num_types, sizeof(unsigned long) );
		
		if (!SKIP_TO_BLOCK)
		{
			fprintf (stderr, _("Not enough memory\n"));
			exit(1);
		}
	}

	for (type=0; type<num_types; type++)
	{
		ind_type = type;
		int text = *texts[type];

		if (copy_file_template_frags_list)
		{
			file_frags_list = copy_file_template_frags_list;
			copy_file_template_frags_list = NULL;
		}
		else
			copy_frags_list (file_template_frags_list, &file_frags_list);

		blocks_before_frag = 0;
		cur_frag = file_frags_list;
		fd_seek(0, SEEK_SET);
		if (!text || (get_block()>=SKIP_TO_BLOCK[type]) ) {
			mes = surrects[type](); 
			if (mes) { 
				mode_t mode = *modes[type];
				char* dupmes = strdup(mes);
				max_size = max_t(any_size_t, max_size, fd_seek(0, SEEK_CUR) );
				anysurrect_file(info, dupmes, mode);
				free(dupmes);
				free_frags_list (file_frags_list);
			}
			else {	
				any_size_t pos = fd_seek(0, SEEK_CUR);
				if (text) SKIP_TO_BLOCK[type] = get_block() + 
					((pos-1) >> get_log2blocksize()) + 1;
				copy_file_template_frags_list = file_frags_list;
			}
		}
		else copy_file_template_frags_list = file_frags_list;
	}

	ind_type = -1;

	set_block( ( (max_size-1) >> get_log2blocksize() ) + 1 );
}

void MainWindow::EmitAddItem(struct any_sb_info *info,            
		const char *name, uint32_t newino, uint32_t dirino)
{
	emit SigAddItem(info, name, newino, dirino);
}

void MainWindow::AddItem(struct any_sb_info *info,            
		const char *name, uint32_t newino, uint32_t dirino)
{
	if (!state) 
	{
		printf ("state = 0\n");
		return; 
	}
	QTreeWidgetItem *item;
	QTreeWidgetItem *parent;
	struct any_inode *dir_inode;
        
	item = mainwindow->GetItem(info, name, newino);
	dir_inode = info->si_inode_table + dirino;
	parent = (QTreeWidgetItem *)
		dir_inode->i_info.dir->d_data;
	assert(parent);
	parent->addChild(item);

	free( (void*) name );
}

int MainWindow::FillItems(uint32_t ino, char *name, QTreeWidgetItem *parent)
{
	struct any_sb_info *info = sb_info;

	struct any_dirent* dirent;

	QTreeWidgetItem *item = GetItem(info, name, ino);
	if (!parent)
		treeWidget->insertTopLevelItem(0, item);
	else
		parent->addChild(item);

	struct any_inode *inode = info->si_inode_table + ino;
	if ( inode->i_mode>>12 == 04 )
	{
		dirent = info->si_inode_table[ino].i_info.dir->d_dirent;

		for(; dirent; dirent = dirent->d_next)
		{
			FillItems(dirent->d_inode, dirent->d_name, item);
			if (!open_state) return 1;
		}
	}

	QCoreApplication::processEvents();

	if (!open_state) return 1;

	return 0;
}

QTreeWidgetItem *MainWindow::GetItem(struct any_sb_info *info,
		const char *name, uint32_t ino)
{
	QFileIconProvider ip;
	int icontype = -1;

	QStringList columnData;
	struct any_inode *inode = info->si_inode_table + ino;
	columnData << QString(name);
	QString str; 
	str.setNum(inode->i_size); columnData << str;

	switch(inode->i_mode>>12)
	{
		case 01:
			columnData << _("FIFO");
			break;
		case 02:
			columnData << _("Character Device");
			break;
		case 04:
			columnData << _("Directory");
			icontype = QFileIconProvider::Folder;
			break;
		case 06:
			columnData << _("Block Device");
			icontype = QFileIconProvider::Drive;
			break;
		case 010:
			columnData << _("Regular File");
			icontype = QFileIconProvider::File;
			break;
		case 012:
			columnData << _("Symbolic Link");
			break;
		case 014:
			columnData << _("Socket");
			break;
		default:
			columnData << _("Unknown");
			break;
	}

	if (ino==1)
		icontype = QFileIconProvider::Drive;

	if ( inode->i_mode>>12 == 012 )
		columnData << inode->i_info.symlink;
	else if ( inode->i_mode>>12 != 010 &&
	       inode->i_mode>>12 != 04 )
	{
		str.setNum(inode->i_info.device); 
		columnData << str;
	}
	else
		columnData << "N/A";

	str.setNum(inode->i_mode & 0xFFF, 8); columnData << str;
	str.setNum(inode->i_uid); columnData << str;
	str.setNum(inode->i_gid); columnData << str;
	char timestr[100];
	timestr[0]='\0'; strftime(timestr, 100, "%c",
			localtime((const time_t*)&inode->i_atime));
	columnData << QString::fromUtf8(timestr);
	timestr[0]='\0'; strftime(timestr, 100, "%c",
			localtime((const time_t*)&inode->i_ctime));
	columnData << QString::fromUtf8(timestr);
	timestr[0]='\0'; strftime(timestr, 100, "%c",
			localtime((const time_t*)&inode->i_mtime));
	columnData << QString::fromUtf8(timestr);
	str.setNum(inode->i_links_count); columnData << str;
	str.setNum(ino); columnData << str;

	QTreeWidgetItem *item = new QTreeWidgetItem(columnData);
	if (icontype>=0)
	item->setIcon( 0, ip.icon(QFileIconProvider::IconType(icontype)) );

	if ( inode->i_mode>>12 == 04 )
		inode->i_info.dir->d_data = (void *) item;

	return item;
}

int MainWindow::SaveInode(const QString &To,
		uint32_t ino)
{
/*
------------------------------------8<----------------------------------
--------------------------->8-------------------------------------------
 */
}

int MainWindow::SaveSelected()
{
	device_name = strdup(diskBox->currentText().toAscii().data());

	QTreeWidgetItem *item = treeWidget->currentItem();
	uint32_t ino = item->text(11).toInt();

	QFileDialog fd(this);

	QString FileName = fd.getSaveFileName();
	if ( FileName.isNull() )
		return 1;

	return SaveInode(FileName, ino);
}

int MainWindow::InitSurrect()
{
/*
------------------------------------8<----------------------------------
--------------------------->8-------------------------------------------
 */
}

int MainWindow::DeInitSurrect()
{
	int lib;

	for (lib=0; lib<num_libs; lib++)
	{
		if ( DLCLOSE(libs_handles[lib]) )
		{
			fprintf( stderr, "Error dlclose(): %s\n",
					dlerror() );
			exit(1);
		}
	}

	anysurrect_free_clean();
}

void MainWindow::Message(QString str1, QString str2)
{
	QMessageBox::critical(this, str1, str2);
}

int MainWindow::StartSurrect()
{
/*
------------------------------------8<----------------------------------
--------------------------->8-------------------------------------------
 */
}

void MainWindow::FinishSurrect()
{
	if (thread)
	{
		if ( !thread->isFinished() )
		{
			wait_terminate = 1;
			thread->wait();
			wait_terminate = 0;
		}

		delete thread;
		thread = NULL;
	}
	
	anysurrect_free_clean();
	if (block_bitmap) free(block_bitmap);
	block_bitmap = NULL;
	if (FD!=-1)
		CLOSE(FD);
	FD = -1;
	fd = FD;
	state = 2;
}

int MainWindow::NewSurrect()
{
	if ( state && (state==1 || !saved) )
	{
		if ( !maybeSave() )
			return 1;
	}

	FinishSurrect();
	if (sb_info)
		free_it (sb_info);
	sb_info = NULL;
	treeWidget->clear();
	state = 0;

	blocksizeBox->setEnabled(true);
	firstblockBox->setEnabled(true);

	actionNew->setEnabled(0);
	actionSave->setEnabled(0);
	actionSave_As->setEnabled(0);

	saved = 0;

	open_state = 0;

	return 0;
}

int MainWindow::save()
{
	int r = 0;

	if ( state == 1 && !thread->isFinished() ) 
	{
		save_state = 1;
		while ( save_state < 2 && !thread->isFinished() ) 
#if defined(__linux__)
			usleep(1000);
#else
			Sleep(1);
#endif
	}

	r = write_it (sb_info, NULL);

	save_state = 0;

	if (r)
	{
		fprintf(stderr,
				_("Error while writing inode table: %s\n"),
				(errno)?strerror(errno):_("format error"));
		QMessageBox::critical(this, _("Write Error"), 
				_("Error while writing inode table."));
		return r;
	}

	if ( state == 2 )
	{
		saved = 1;
		actionSave->setEnabled(0);
		actionSave_As->setEnabled(0);
	}

	return 0;
}

int MainWindow::saveAction()
{
	if (!sb_info) return 1;
	if (!sb_info->si_itfilename)
	{
		QFileDialog fd(this);

		QString FileName = fd.getSaveFileName();
		if ( FileName.isNull() )
			return 1;

		sb_info->si_itfilename = strdup(FileName.toAscii().data());
	}

	return save();
}

int MainWindow::saveAsAction()
{
	if (!sb_info) return 1;
	QFileDialog fd(this);

	QString FileName = fd.getSaveFileName();
	if ( FileName.isNull() )
		return 1;

	sb_info->si_itfilename = strdup(FileName.toAscii().data());

	return save();
}

int MainWindow::openAction()
{
	int r;

	QFileDialog fd(this);
	QString FileName = fd.getOpenFileName();
	if ( FileName.isNull() )
		return 1;

	if ( NewSurrect() )
		return 1;

	char *input_inode_table = strdup(FileName.toAscii().data());

	r = read_it (&sb_info, (char*) input_inode_table);
	if (r)
	{
		fprintf(stderr,
				_("Error while reading inode table: %s\n"),
				(errno)?strerror(errno):_("bad format"));
		QMessageBox::critical(this, _("Open Error"), 
				_("Error while reading inode table."));
		return r;
	}

	set_blocksize(sb_info->si_blocksize);
	
	QString str;
	str.setNum(sb_info->si_blocksize);
	blocksizeBox->setEditText(str);

	blocksizeBox->setEnabled(false);
	firstblockBox->setEnabled(false);

	state = 2;
	saved = 1;

	actionNew->setEnabled(1);
	actionSave->setEnabled(0);
	actionSave_As->setEnabled(0);

	checkState();

	open_state = 1;

	FillItems();
}

int MainWindow::newAction()
{
	if ( ! NewSurrect() )
		checkState();
}

